""" Twempest render logic unit tests.
"""

# This file is part of Twempest. Copyright 2017 Dave Rogers <info@yukondude.com>. Licensed under the GNU General Public License, version 3.
# Refer to the attached LICENSE file or see <http://www.gnu.org/licenses/> for details.

import glob
import os
import pickle
import re

from click import echo
from click.testing import CliRunner
import jinja2
# noinspection PyPackageRequirements
import pytest

from twempest.twempest import PICKLE_FILE_NAME, cleanup_downloaded_images, download_from_url, download_images, render, TwempestError
from twempest.__main__ import CONFIG_OPTIONS
from .fixtures import MockEcho, tweets_fixture


IMAGE_TWEET_IDS = (806229878861201408, 810533832529219584, 812831603051220992, 814688934835826688)
REPLY_TWEET_IDS = (805892124004618240, 812686867522953216)
TEST_IMAGE_URL = "http://placehold.it/10x10.jpg"
TEST_IMAGE_SIZE = 3881


def mock_download(url, file_path):
    """ Write a ~100KB file to file_path as if it were downloaded.
    """
    with open(file_path, "wb") as f:
        f.write(url.encode('utf-8'))
        f.seek(1024 * 1024)
        f.write(b'0')


@pytest.fixture
def mock_echo():
    return MockEcho()


@pytest.fixture
def tweets():
    return tweets_fixture()


# noinspection PyShadowingNames
def test_cleanup_downloaded_images(mock_echo):
    with CliRunner().isolated_filesystem():
        image_file_paths = []

        for i in range(4):
            image_file_name = "image{}.jpg".format(i)
            image_file_path = os.path.join(".", image_file_name)
            image_file_paths.append(image_file_path)
            with open(image_file_path, 'w') as f:
                f.write("foo")

        cleanup_downloaded_images(image_file_paths, echo)
        assert len(mock_echo.errs) == len(mock_echo.messages) == 0

        for path in image_file_paths:
            assert not os.path.exists(path)


# noinspection PyShadowingNames
def test_cleanup_downloaded_images_fail_cant_delete(mock_echo):
    with CliRunner().isolated_filesystem():
        image_file_path = os.path.join(".", "quux.jpg")
        cleanup_downloaded_images([image_file_path], mock_echo.echo)
        assert len(mock_echo.errs) == len(mock_echo.messages) == 1
        assert mock_echo.errs[0]
        assert mock_echo.messages[0].startswith("Warning: Unable to delete downloaded image file:")
        assert image_file_path in mock_echo.messages[0]


def test_download_from_url():
    with CliRunner().isolated_filesystem():
        path = os.path.join("download", "test.jpg")
        os.mkdir("download")
        download_from_url(TEST_IMAGE_URL, path)
        assert os.path.exists(path)
        stat_info = os.stat(path)
        assert stat_info.st_size == TEST_IMAGE_SIZE


def test_download_from_url_fail_url():
    with CliRunner().isolated_filesystem():
        path = os.path.join("download", "test.jpg")
        os.mkdir("download")

        with pytest.raises(TwempestError) as excinfo:
            download_from_url("http://test.com/doesntexist", path)

        assert "Unable to download image file:" in str(excinfo.value)
        assert not os.path.exists(path)


def test_download_from_url_fail_save():
    with CliRunner().isolated_filesystem():
        path = os.path.join("download", "test.jpg")
        os.mkdir("download", 0o111)

        with pytest.raises(TwempestError) as excinfo:
            download_from_url(TEST_IMAGE_URL, path)

        assert "Unable to write downloaded image file:" in str(excinfo.value)
        assert not os.path.exists(path)


# noinspection PyShadowingNames
def test_download_images(mock_echo, tweets):
    env = jinja2.Environment()
    image_dir_path_template = env.from_string("images")
    image_url_path_template = env.from_string("/images/")
    image_paths = []

    with CliRunner().isolated_filesystem():
        for tweet in tweets:
            render_file_name = "tweet_{}_file.md".format(tweet.id)
            paths = download_images(tweet, image_dir_path_template, image_url_path_template, render_file_name, mock_download,
                                    mock_echo.echo)

            if paths:
                image_paths.append(paths[0])

        assert len(image_paths) == len(IMAGE_TWEET_IDS)

        for tweet_id in IMAGE_TWEET_IDS:
            path = os.path.join("images", "tweet_{}_file-0.jpg".format(tweet_id))
            assert os.path.exists(path)
            stat_info = os.stat(path)
            assert stat_info.st_size > 1024 * 1024
            assert len([1 for p in image_paths if p.endswith(path)]) == 1, "path '{}' not found in image_paths list.".format(path)


# noinspection PyShadowingNames
def test_download_images_some_exist(mock_echo, tweets):
    env = jinja2.Environment()
    image_dir_path_template = env.from_string("images")
    image_url_path_template = env.from_string("/images/")
    image_paths = []

    with CliRunner().isolated_filesystem():
        for tweet in tweets:
            render_file_name = "tweet_{}_file.md".format(tweet.id)

            if tweet.id == IMAGE_TWEET_IDS[-1]:
                with open(os.path.join("images", "tweet_{}_file-0.jpg".format(tweet.id)), "wb") as f:
                    f.seek(1024 * 1024)
                    f.write(b'0')

            paths = download_images(tweet, image_dir_path_template, image_url_path_template, render_file_name, mock_download,
                                    mock_echo.echo)

            if paths:
                image_paths.append(paths[0])

            if tweet.id == IMAGE_TWEET_IDS[-1]:
                assert len(mock_echo.messages) == 1
                assert "Warning: Skipping existing image file" in mock_echo.messages[0]
                assert str(tweet.id) in mock_echo.messages[0]

        assert len(image_paths) == len(IMAGE_TWEET_IDS) - 1


# noinspection PyShadowingNames
def test_download_images_media_update(mock_echo, tweets):
    env = jinja2.Environment()
    image_dir_path_template = env.from_string("images")
    image_url_path_template = env.from_string("/images/")
    media_before = {}
    media_after = {}

    with CliRunner().isolated_filesystem():
        for tweet in tweets:
            render_file_name = "tweet_{}_file.md".format(tweet.id)
            media_before[tweet.id] = [{'media_url': m['media_url'],
                                       'media_url_https': m['media_url_https']}
                                      for m in tweet.entities.get('media', []) if m['type'] == 'photo']

            download_images(tweet, image_dir_path_template, image_url_path_template, render_file_name, mock_download, mock_echo.echo)

            media_after[tweet.id] = [{'media_url': m['media_url'],
                                      'media_url_https': m['media_url_https'],
                                      'original_media_url': m['original_media_url'],
                                      'original_media_url_https': m['original_media_url_https']}
                                     for m in tweet.entities.get('media', []) if m['type'] == 'photo']

        for tweet_id in IMAGE_TWEET_IDS:
            before = media_before[tweet_id]
            after = media_after[tweet_id]

            for i, b in enumerate(before):
                a = after[i]
                assert b['media_url'] == a['original_media_url']
                assert b['media_url'] != a['media_url']
                assert b['media_url_https'] == a['original_media_url_https']
                assert b['media_url_https'] != a['media_url_https']


# noinspection PyShadowingNames
def test_render_to_console(mock_echo, tweets):
    with CliRunner().isolated_filesystem():
        template_text = "{{ tweet.id }}"
        options = {k: v.default for k, v in CONFIG_OPTIONS.items()}
        options['replies'] = True
        last_id = render(tweets, options, template_text, mock_download, mock_echo.echo)
        rendered_ids = [i for i in mock_echo.messages if i]
        assert len(rendered_ids) == len(tweets)
        assert str(last_id) == rendered_ids[-1]

        for tweet in tweets:
            assert str(tweet.id) in rendered_ids


# noinspection PyShadowingNames
def test_render_to_files(mock_echo, tweets):
    with CliRunner().isolated_filesystem():
        template_text = "{{ tweet.id }}"
        options = {k: v.default for k, v in CONFIG_OPTIONS.items()}
        options['render-file'] = "{{ tweet.id }}.txt"
        options['replies'] = True
        last_id = render(tweets, options, template_text, mock_download, mock_echo.echo)
        assert len(mock_echo.messages) == 0
        text_file_names = glob.glob("*.txt")
        assert len(tweets) == len(text_file_names)
        assert "{}.txt".format(last_id) == text_file_names[-1]

        for tweet in tweets:
            file_name = "{}.txt".format(tweet.id)
            assert os.path.exists(file_name)
            with open(file_name, "r") as f:
                assert str(tweet.id) == f.read()


# noinspection PyShadowingNames
def test_render_no_replies(mock_echo, tweets):
    with CliRunner().isolated_filesystem():
        template_text = "{{ tweet.id }}"
        options = {k: v.default for k, v in CONFIG_OPTIONS.items()}
        last_id = render(tweets, options, template_text, mock_download, mock_echo.echo)
        rendered_ids = [i for i in mock_echo.messages if i]
        assert len(rendered_ids) == len(tweets) - len(REPLY_TWEET_IDS)
        assert str(last_id) == rendered_ids[-1]

        for rid in REPLY_TWEET_IDS:
            assert str(rid) not in rendered_ids


# noinspection PyShadowingNames
def test_render_no_append(mock_echo, tweets):
    with CliRunner().isolated_filesystem():
        template_text = "{{ tweet.id }}"
        options = {k: v.default for k, v in CONFIG_OPTIONS.items()}
        options['count'] = 2
        options['render-file'] = "tweet.txt"
        render(tweets, options, template_text, mock_download, mock_echo.echo)
        assert len(mock_echo.messages) == 1
        assert "Warning: Skipping existing file '" in mock_echo.messages[0]
        assert "/tweet.txt'. Use --append" in mock_echo.messages[0]

        with open("tweet.txt", "r") as f:
            assert str(tweets[0].id) == f.read()


# noinspection PyShadowingNames
def test_render_with_append(mock_echo, tweets):
    with CliRunner().isolated_filesystem():
        template_text = "{{ tweet.id }}"
        options = {k: v.default for k, v in CONFIG_OPTIONS.items()}
        options['append'] = True
        options['count'] = 2
        options['render-file'] = "tweet.txt"
        render(tweets, options, template_text, mock_download, mock_echo.echo)
        assert len(mock_echo.messages) == 0

        with open("tweet.txt", "r") as f:
            assert str(tweets[0].id) + str(tweets[1].id) == f.read()


# noinspection PyShadowingNames
def test_render_with_skip(mock_echo, tweets):
    with CliRunner().isolated_filesystem():
        template_text = "{{ tweet.id }}"
        options = {k: v.default for k, v in CONFIG_OPTIONS.items()}
        options['skip'] = re.compile("806229878861201408")
        options['render-file'] = "{{ tweet.id }}.txt"
        options['replies'] = True
        render(tweets, options, template_text, mock_download, mock_echo.echo)
        assert len(mock_echo.messages) == 1
        assert "Warning: Skipping tweet ID 806229878861201408 ('" in mock_echo.messages[0]
        text_file_names = glob.glob("*.txt")
        assert len(tweets) - 1 == len(text_file_names)


# noinspection PyShadowingNames
def test_render_skip_all(mock_echo, tweets):
    with CliRunner().isolated_filesystem():
        template_text = "{{ tweet.id }}"
        options = {k: v.default for k, v in CONFIG_OPTIONS.items()}
        options['replies'] = True
        options['skip'] = re.compile(".")
        last_id = render(tweets, options, template_text, mock_download, mock_echo.echo)
        assert last_id is None
        assert len(mock_echo.messages) == len(tweets) + 1
        assert "Warning: No tweets were retrieved." == mock_echo.messages[-1]


# noinspection PyShadowingNames
def test_render_pickle(mock_echo, tweets):
    with CliRunner().isolated_filesystem():
        template_text = "{{ tweet.id }}"
        options = {k: v.default for k, v in CONFIG_OPTIONS.items()}
        options['count'] = 5
        options['pickle'] = True
        options['replies'] = True
        render(tweets, options, template_text, mock_download, mock_echo.echo)

        assert os.path.exists(PICKLE_FILE_NAME)
        pickled_tweets = pickle.load(open(PICKLE_FILE_NAME, "rb"))
        assert len(pickled_tweets) == 5

        for i, pickled_tweet in enumerate(pickled_tweets):
            assert pickled_tweet.id == tweets[i].id



""" Twempest unit test fixtures.
"""

# This file is part of Twempest. Copyright 2017 Dave Rogers <info@yukondude.com>. Licensed under the GNU General Public License, version 3.
# Refer to the attached LICENSE file or see <http://www.gnu.org/licenses/> for details.

import os
import pickle


class MockEcho:
    """ Mock class for testing output of echo() calls.
    """
    messages = []
    errs = []

    def __init__(self):
        del MockEcho.messages[:]
        del MockEcho.errs[:]

    @staticmethod
    def echo(message="", err=False):
        MockEcho.messages.append(message)
        MockEcho.errs.append(err)


def tweets_fixture():
    """ Just a whole mess--well, eighteen--of pickled sample tweets from late 2016 to test with.
    """
    pickle_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "twempest.p")
    return pickle.load(open(pickle_path, "rb"))



""" Twempest custom template filters unit tests.
"""

# This file is part of Twempest. Copyright 2017 Dave Rogers <info@yukondude.com>. Licensed under the GNU General Public License, version 3.
# Refer to the attached LICENSE file or see <http://www.gnu.org/licenses/> for details.

import datetime
import os
import re

# noinspection PyPackageRequirements
import pytest

from twempest.filters import delink, isodate, reimage, relink, slugify
from .fixtures import tweets_fixture


class MockContext:
    """ Stand-in for Jinja template context object.
    """
    def __init__(self, tweet):
        self.parent = {'tweet': tweet}


@pytest.fixture
def tweets():
    return tweets_fixture()


# noinspection PyShadowingNames
def test_delink(tweets):
    for tweet in tweets:
        delinked = delink(ctx=MockContext(tweet), text=tweet.text)

        for hashtag in tweet.entities.get('hashtags', []):
            assert '#' + hashtag['text'] not in delinked

        for media in tweet.entities.get('media', []):
            assert media['url'] not in delinked

        for url in tweet.entities.get('urls', []):
            assert url['url'] not in delinked


@pytest.mark.parametrize('date,expected', [
    (datetime.datetime(2017, 2, 14, 12, 44, 57, 557000), "2017-02-14"),
    (datetime.datetime(2016, 1, 1, 14, 23, 44, 590), "2016-01-01"),
    (datetime.date(2016, 2, 29), "2016-02-29"),
    (datetime.date(2016, 12, 31), "2016-12-31"),
])
def test_isodate(date, expected):
    assert isodate(date) == expected


# noinspection PyShadowingNames
def test_reimage(tweets):
    for tweet in tweets:
        reimaged = reimage(ctx=MockContext(tweet), text=tweet.text, tag_format="@@{{alt}}@@{{url}}@@")

        for media in tweet.entities.get('media', []):
            image_url = media['media_url_https'] if media['media_url_https'] else media['media_url']
            image_alt = os.path.splitext(os.path.basename(image_url))[0]

            if media['type'] == 'photo':
                assert media['url'] not in reimaged
                assert "@@{}@@{}@@".format(image_alt, image_url) in reimaged
            else:
                assert media['url'] in reimaged
                assert "@@{}@@{}@@".format(image_alt, image_url) not in reimaged

        for url in tweet.entities.get('urls', []):
            assert url['url'] in reimaged
            assert "@@{}@@{}@@".format(url['display_url'], url['expanded_url']) not in reimaged


# noinspection PyShadowingNames
def test_relink(tweets):
    for tweet in tweets:
        relinked = relink(ctx=MockContext(tweet), text=tweet.text, tag_format="@@{{text}}@@{{url}}@@")

        for hashtag in tweet.entities.get('hashtags', []):
            assert "@@#{}@@https://twitter.com/hashtag/{}@@".format(hashtag['text'], hashtag['text'].lower()) in relinked

        for media in tweet.entities.get('media', []):
            if media['type'] == 'photo':
                assert media['url'] in relinked
                assert "@@{}@@{}@@".format(media['display_url'], media['expanded_url']) not in relinked
            else:
                assert media['url'] not in relinked
                assert "@@{}@@{}@@".format(media['display_url'], media['expanded_url']) in relinked

        for url in tweet.entities.get('urls', []):
            assert url['url'] not in relinked
            assert "@@{}@@{}@@".format(url['display_url'], url['expanded_url']) in relinked


# noinspection PyShadowingNames
def test_slugify(tweets):
    only_allowed = re.compile(r"^[a-z0-9-]+$")

    for tweet in tweets:
        slugified = slugify(ctx=MockContext(tweet), text=tweet.text)
        assert only_allowed.match(slugified)
        assert "--" not in slugified
        assert not slugified.startswith('-')
        assert not slugified.endswith('-')



""" Twitter to text via template. Somehow that abbreviates to "twempest".
"""

# This file is part of Twempest.
#
# Twempest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your option) any later version.
#
# Twempest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Twempest. If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2017 Dave Rogers <info@yukondude.com>

__version__ = "0.2.5"



""" Twempest console script entry point.
"""

# This file is part of Twempest. Copyright 2017 Dave Rogers <info@yukondude.com>. Licensed under the GNU General Public License, version 3.
# Refer to the attached LICENSE file or see <http://www.gnu.org/licenses/> for details.

import collections
import configparser
import hashlib
import os
import re

import click

from .twempest import retrieve, TwempestError


# Global config 'constants'.
CONFIG_FILE_NAME = "twempest.config"
DEFAULT_CONFIG_DIR_PATH = "~/.twempest"
FALLBACK_CONFIG_DIR_PATH = "."

# Collect all configuration options (that may also appear in the config file) here so that they don't have to be duplicated.
ConfigOption = collections.namedtuple('ConfigOption', "short default show_default is_flag help")
CONFIG_OPTIONS = {
    'append': ConfigOption(short='a', default=False, show_default=False, is_flag=True,
                           help="Append rendered tweet(s) to existing file(s) rather than skipping past with a warning."),
    'count': ConfigOption('n', 200, True, False, "Maximum number of tweets to retrieve. The actual number may be lower."),
    'dry-run': ConfigOption('D', False, False, True, "Display all configuration options and template contents without retrieving tweets."),
    'image-path': ConfigOption('i', None, False, False, "The directory path (template tags allowed) to write downloaded image "
                                                        "(media type == 'photo') files. "
                                                        "The directory path will be created if it doesn't exist. "
                                                        "Media file names use the --render-file name followed by a number and the "
                                                        "appropriate file extension. "
                                                        "If omitted, media files will not be downloaded."),
    'image-url': ConfigOption('u', None, False, False, "The URL path (template tags allowed) to use for all image files downloaded via the "
                                                       "--image-path option."),
    'pickle': ConfigOption(None, False, False, True, "Serialize a list of the rendered tweet statuses as a standard Python pickle byte "
                                                     "stream. The stream will be written to 'twempest.p' in the current working "
                                                     "directory."),
    'render-file': ConfigOption('f', None, False, False, "The file name (template tags allowed) for the rendered tweets. "
                                                         "If omitted, tweets will be rendered to STDOUT."),
    'render-path': ConfigOption('p', ".", True, False, "The directory path (template tags allowed) to write the rendered tweet files. "
                                                       "The directory path will be created if it doesn't exist."),
    'replies': ConfigOption('@', False, False, True, "Include @replies in the list of retrieved tweets."),
    'retweets': ConfigOption('r', False, False, True, "Include retweets in the list of retrieved tweets."),
    'since-id': ConfigOption('s', None, False, False, "Retrieve tweets that follow this ID in the timeline. "
                                                      "Required, unless the ID has already been recorded in the config path directory "
                                                      "after a previous run of Twempest."),
    'skip': ConfigOption('k', None, False, False, "Skip any rendered tweets that contain this regular expression pattern."),
}


def choose_config_path(cli_dir_path, default_dir_path, fallback_dir_path, file_name):
    """ Choose the most likely configuration path from, in order: the CLI config-path option, the default path, and the fallback path
        (current directory). To be valid, the config path must contain a twempest.conf file and the directory must be writable. Return
        both the chosen path and the possible paths as a tuple.
    """
    # Using keys of OrderedDict simulates an OrderedSet type.
    unique_paths = collections.OrderedDict([(os.path.abspath(os.path.expanduser(p)), None) for p in (cli_dir_path, default_dir_path,
                                                                                                     fallback_dir_path)])
    possible_paths = list(unique_paths.keys())

    for possible_path in possible_paths:
        possible_config_path = os.path.join(possible_path, file_name)

        if os.path.isfile(possible_config_path) and os.access(possible_config_path, os.R_OK) and os.access(possible_path, os.W_OK):
            return possible_path, possible_paths

    return None, possible_paths


def choose_option_values(config_options, cli_options, config):
    """ For each of the options given, choose a value from, in order: the CLI option, the config file, the CONFIG_OPTIONS default. Return
        the option values in a tuple.
    """
    options = {}

    for option, config_option in config_options.items():
        cli_option = option.replace('-', '_')
        config_func = config.getboolean if config_option.is_flag else config.get

        if (config_option.is_flag and not cli_options.get(cli_option, False)) or cli_options.get(cli_option) is None:
            # Ignore False CLI flags or unset CLI options so that they don't mask config file or option defaults.
            options[option] = config_func(option, config_option.default)
        else:
            options[option] = cli_options.get(cli_option)

    return options


def choose_since_id(cli_since_id, user_id, config_dir_path):
    """ Choose the most likely since ID value from, in order: the CLI since-id option, and a record of the ID written to a file in the
        config directory using the given user_id to distinguish it from others.
    """
    since_id = cli_since_id

    if since_id:
        return since_id

    # Hash the user ID so that it's unique to that user, but doesn't reveal part of the authentication credentials in the file name.
    last_id_file_name = last_tweet_id_file_name(user_id)
    last_id_file_path = os.path.join(config_dir_path, last_id_file_name)

    if os.path.isfile(last_id_file_path):
        with open(last_id_file_path, 'r') as f:
            since_id = f.read()

    return since_id


def decorate_config_options(options):
    """ Return a decorator with all of the CONFIG_OPTIONS items as a chain of click.option decorators, sorted by option name.
    """
    def option_decorators(fn):
        """ Return the decorator chain wrapped around the given function.
        """
        decorators = fn

        for option, config_option in sorted(options.items(), reverse=True):
            help_text = config_option.help

            if config_option.show_default:
                help_text += "  [default: {}]".format(config_option.default)

            # Don't specify a default value for the option, as it will be the fallback when choose_option_values() is called. For this
            # reason, append the default to the help text manually if it is to be shown.
            args = ["--" + option]

            if config_option.short:
                args.append("-" + config_option.short)

            kwargs = {'is_flag': config_option.is_flag, 'help': help_text}
            decorator = click.option(*args, **kwargs)
            decorators = decorator(decorators)

        return decorators

    return option_decorators


def last_tweet_id_file_name(user_id):
    """ Return the last tweet ID file name corresponding to the given user.
    """
    return "twempest-last-{}.id".format(hashlib.sha1(user_id.encode('utf-8')).hexdigest())


# noinspection PyUnusedLocal
def show_version(ctx, param, value):
    """ Display the version message.
    """
    if not value or ctx.resilient_parsing:
        return  # pragma: no cover
    from twempest import __version__
    click.echo("Twempest version {}".format(__version__))
    click.echo("Copyright 2017 Dave Rogers. Licensed under the GPLv3. See LICENSE.")
    ctx.exit()


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.option("--config-path", "-c", default=DEFAULT_CONFIG_DIR_PATH, show_default=True,
              help="Twempest configuration directory path, which must be writable, and must also contain the twempest.conf file.")
@decorate_config_options(CONFIG_OPTIONS)
@click.option("--version", "-V", is_flag=True, callback=show_version, expose_value=False, is_eager=True, help="Show version and exit.")
@click.argument("template", type=click.File('r'))
def twempest(**kwargs):
    """ Download a sequence of recent Twitter tweets and convert these, via the given template file, to text format.
        Twempest uses the Jinja template syntax throughout: http://jinja.pocoo.org/docs/2.9/templates/
    """
    config = configparser.RawConfigParser(allow_no_value=True)
    config_dir_path, possible_paths = choose_config_path(cli_dir_path=kwargs['config_path'], default_dir_path=DEFAULT_CONFIG_DIR_PATH,
                                                         fallback_dir_path=FALLBACK_CONFIG_DIR_PATH, file_name=CONFIG_FILE_NAME)

    if not config_dir_path:
        raise click.ClickException("Could not find readable twempest.conf configuration file in writable directory path(s): '{}'"
                                   .format("', '".join(possible_paths)))

    config_file_path = os.path.join(config_dir_path, CONFIG_FILE_NAME)
    config.read(config_file_path)

    try:
        twitter_config = config['twitter']
        twempest_config = config['twempest']
    except KeyError as e:
        raise click.ClickException("Could not find required '[{}]' section in '{}'".format(str(e).strip("'"), config_file_path))

    try:
        auth_keys = {k: twitter_config[k] for k in ('consumer_key', 'consumer_secret', 'access_token', 'access_token_secret')}
    except KeyError as e:
        raise click.ClickException("Could not find required Twitter authentication credential {} in '{}'".format(str(e), config_file_path))

    options = choose_option_values(config_options=CONFIG_OPTIONS, cli_options=kwargs, config=twempest_config)
    options['config-path'] = config_file_path

    try:
        options['count'] = int(options['count'])
    except ValueError as e:
        raise click.ClickException("The --count option must be an integer: {}".format(e))

    if options['count'] < 1:
        raise click.ClickException("The --count option must be at least 1.")

    if options['image-path'] and not options['render-file']:
        raise click.ClickException("Cannot download images unless the --render-file option is also specified.")

    if options['image-url'] and not options['image-path']:
        raise click.ClickException("The --image-url option may only be specified if the --image-path option is as well.")

    options['since-id'] = choose_since_id(cli_since_id=options['since-id'], user_id=twitter_config['consumer_key'],
                                          config_dir_path=config_dir_path)

    if not options['since-id']:
        raise click.ClickException("The --since-id option is required since the ID was not recorded in '{}' after a previous run of "
                                   "Twempest. To find the ID, open a specific tweet on the Twitter website and view the page's address: "
                                   "the long number following 'status/' is that tweet's ID."
                                   .format(config_dir_path))

    if options['skip'] is not None:
        try:
            # Replace skip option with compiled regular expression.
            options['skip'] = re.compile(options['skip'])
        except re.error as e:
            raise click.ClickException("Syntax problem with --skip regular expression: {}".format(e))

    template_file = kwargs['template']

    try:
        template_text = template_file.read()
    except OSError as e:
        raise click.ClickException("Unable to read template file: {}".format(e))

    if options['dry-run']:
        for option in sorted(options.keys()):
            click.echo("{} = {}".format(option, options[option]))

        click.echo()
        click.echo("template =")
        click.echo(template_text)
    else:
        try:
            last_tweet_id = retrieve(auth_keys=auth_keys, options=options, template_text=template_text, echo=click.echo)
        except TwempestError as e:
            raise click.ClickException(e)

        if last_tweet_id:
            try:
                with open(os.path.join(config_dir_path, last_tweet_id_file_name(user_id=twitter_config['consumer_key'])), 'w') as f:
                    f.write(str(last_tweet_id))
            except OSError as e:
                raise click.ClickException("Unable to write last tweet ID file: {}".format(e))



""" Twempest Jinja2 custom filters.
"""

# This file is part of Twempest. Copyright 2017 Dave Rogers <info@yukondude.com>. Licensed under the GNU General Public License, version 3.
# Refer to the attached LICENSE file or see <http://www.gnu.org/licenses/> for details.

import inspect
import os
import re
import unicodedata
import sys

import jinja2


@jinja2.contextfilter
def delink(ctx, text):
    """ Remove URLs and hashtag '#' prefixes from the given text, using the contents of the context's tweet status object as a guide.
    """
    tweet_entities = ctx.parent['tweet'].entities

    for hashtag in tweet_entities.get('hashtags', []):
        text = text.replace('#' + hashtag['text'], hashtag['text'])

    for media in tweet_entities.get('media', []):
        text = text.replace(media['url'], '')

    for url in tweet_entities.get('urls', []):
        text = text.replace(url['url'], '')

    return text


def isodate(date):
    """ Return the given date formatted as the ISO 8601 extended YYYY-MM-DD format.
    """
    return date.strftime('%Y-%m-%d')


@jinja2.contextfilter
def reimage(ctx, text, tag_format, delimiter=" "):
    """ Remove image URLs and append them to the end of the given text, following the template tag_format with variables 'alt' and 'url',
        using the context's tweet status object to supply the necessary values.
    """
    tweet_entities = ctx.parent['tweet'].entities
    images = []

    for media in (m for m in tweet_entities.get('media', []) if m['type'] == 'photo'):
        text = text.replace(media['url'], '')
        image_url = media['media_url_https'] if media['media_url_https'] else media['media_url']
        # Use the filename without extension for the alt text since there's really nothing better that's available.
        image_alt = os.path.splitext(os.path.basename(image_url))[0]
        images.append((image_alt, image_url))

    image_template = jinja2.Template(tag_format)
    text = text.rstrip()

    for image_alt, image_url in images:
        text += delimiter + image_template.render(alt=image_alt, url=image_url)

    return text


@jinja2.contextfilter
def relink(ctx, text, tag_format):
    """ Add non-image URLs, hashtag, and user mention links to the given text, following the template tag_format with variables 'text' and
        'url', using the context's tweet status object to supply the necessary values.
    """
    tweet_entities = ctx.parent['tweet'].entities
    link_template = jinja2.Template(tag_format)

    for hashtag in tweet_entities.get('hashtags', []):
        hashtag_text = hashtag['text']
        hashtag_url = "https://twitter.com/hashtag/{}".format(hashtag_text.lower())
        text = text.replace('#' + hashtag_text, link_template.render(text='#' + hashtag_text, url=hashtag_url))

    for url in tweet_entities.get('urls', []):
        text = text.replace(url['url'], link_template.render(text=url['display_url'], url=url['expanded_url']))

    for media in (m for m in tweet_entities.get('media', []) if m['type'] != 'photo'):
        text = text.replace(media['url'], link_template.render(text=url['display_url'], url=url['expanded_url']))

    for mention in tweet_entities.get('user_mentions', []):
        mention_text = mention['screen_name']
        mention_url = "https://twitter.com/{}".format(mention_text)
        text = text.replace('@' + mention_text, link_template.render(text='@' + mention_text, url=mention_url))

    return text


WEIRD_CHARACTERS_RE = re.compile(r"[^\w\s-]")
MULTIPLE_DELIMITERS_RE = re.compile(r"[-\s]+")


@jinja2.contextfilter
def slugify(ctx, text):
    """ Transform the given text into a suitable file name that is also scrubbed of URLs and hashtags.

        This function is adapted from the Django project utils/text.py source code file:
        https://github.com/django/django/blob/master/django/utils/text.py

        Copyright (c) Django Software Foundation and individual contributors. All rights reserved.
        License: https://github.com/django/django/blob/master/LICENSE
    """
    slug = delink(ctx, text)
    slug = unicodedata.normalize('NFKD', slug).encode('ascii', 'ignore').decode('ascii')
    slug = WEIRD_CHARACTERS_RE.sub('', slug).strip().lower()
    return MULTIPLE_DELIMITERS_RE.sub('-', slug).strip('-')


# Dictionary {name: function} of all filter functions in this module.
ALL_FILTERS = {m[0]: m[1] for m in inspect.getmembers(sys.modules[__name__]) if inspect.isfunction(m[1])}



""" Twempest rendering logic.
"""

# This file is part of Twempest. Copyright 2017 Dave Rogers <info@yukondude.com>. Licensed under the GNU General Public License, version 3.
# Refer to the attached LICENSE file or see <http://www.gnu.org/licenses/> for details.

import os
import pickle
import urllib.request as request

import jinja2
import pytz
import tweepy
import tzlocal

from .filters import ALL_FILTERS


PICKLE_FILE_NAME = "twempest.p"


class TwempestError(Exception):
    pass


def authenticate_twitter_api(consumer_key, consumer_secret, access_token, access_token_secret):
    """ Return the Twitter API object for the given authentication credentials.
    """
    auth = tweepy.OAuthHandler(consumer_key, consumer_secret)
    auth.set_access_token(access_token, access_token_secret)
    return tweepy.API(auth)


def cleanup_downloaded_images(downloaded_image_file_paths, echo):
    """ Clean up any downloaded images for the skipped tweet. Clunky, but simpler than avoiding downloading in the first place. Warn of any
        image files that couldn't be deleted using the passed echo() function but continue on regardless.
    """
    for path in downloaded_image_file_paths:
        try:
            os.unlink(path)
        except OSError as e:
            echo("Warning: Unable to delete downloaded image file: {}".format(e), err=True)


def download_from_url(url, file_path):
    """ Download the file at the given url and store it at the given file path. Raise a TwempestError exception if anything goes wrong.
    """
    try:
        request.urlretrieve(url, file_path)
    except request.URLError as e:
        raise TwempestError("Unable to download image file: {}".format(e))
    except OSError as e:
        raise TwempestError("Unable to write downloaded image file: {}".format(e))


def download_images(tweet, image_dir_path_template, image_url_path_template, render_file_name, download_func, echo):
    """ Download any images for the given tweet, storing them in the rendered image directory path template and updating their URLs with the
        rendered image URL path template. Echo any skipped images that already exist to the console using the passed echo() function. Return
        the list of file paths for the downloaded images.
    """
    downloaded_image_file_paths = []
    image_dir_path = os.path.abspath(image_dir_path_template.render(tweet=tweet))
    os.makedirs(image_dir_path, exist_ok=True)

    for i, media in enumerate(m for m in tweet.entities.get('media', []) if m['type'] == 'photo'):
        image_download_url = media['media_url_https'] if media['media_url_https'] else media['media_url']
        image_file_name = "{}-{}{}".format(os.path.splitext(render_file_name)[0], i,
                                           os.path.splitext(os.path.basename(image_download_url))[1])
        image_file_path = os.path.join(image_dir_path, image_file_name)
        image_url_path = image_url_path_template.render(tweet=tweet).rstrip('/') + '/' + image_file_name

        # Inject the downloaded image URL into the tweet status object and backup the original media URL(s) just in case.
        is_https = image_url_path.lower().startswith("https")
        media['original_media_url_https'] = media['media_url_https']
        media['media_url_https'] = image_url_path if is_https else None
        media['original_media_url'] = media['media_url']
        media['media_url'] = image_url_path if not is_https else None

        if os.path.exists(image_file_path):
            echo("Warning: Skipping existing image file '{}'.".format(image_file_path), err=True)
            continue

        download_func(image_download_url, image_file_path)
        downloaded_image_file_paths.append(image_file_path)

    return downloaded_image_file_paths


def render(tweets, options, template_text, download_func, echo):
    """ Render the given tweets using the supplied template text. Also download images if requested. Write any warning messages to the
        console using the passed echo() function, and raise all errors as TwempestError.
    """
    def write_to_console(text):
        """ Write the given text to the console with a following blank line.
        """
        echo(text)
        echo()

    def write_to_file(path):
        """ Return a function that will write text to a file at the given path.
        """
        def write_to_file_inner(text):
            """ Write the given text to the outer function's file path.
            """
            try:
                with open(path, 'a') as f:
                    f.write(text)
            except OSError as oe:
                raise TwempestError("Unable to write rendered tweet: {}".format(oe))

        return write_to_file_inner

    # noinspection PyUnusedLocal
    def write_to_void(text):
        """ Don't do anything with the given text.
        """
        pass

    env = jinja2.Environment()
    env.filters.update(ALL_FILTERS)
    template = env.from_string(template_text)

    image_dir_path_template = env.from_string(options['image-path']) if options['image-path'] else None
    image_url_path_template = env.from_string(options['image-url']) if options['image-url'] else None
    render_file_name_template = env.from_string(options['render-file']) if options['render-file'] else None
    render_dir_path_template = env.from_string(options['render-path'])

    gmt_tz = pytz.timezone('UTC')
    local_tz = tzlocal.get_localzone()

    count_remaining = options['count']
    last_tweet_id = None
    pickle_tweets = []

    for tweet in tweets:
        # Replace UTC created time with local time.
        if tweet.created_at.tzinfo is None:
            tweet.created_at = gmt_tz.localize(tweet.created_at)

        tweet.created_at = tweet.created_at.astimezone(local_tz)

        if not options['replies'] and tweet.in_reply_to_status_id and tweet.text[0] == '@':
            continue

        if render_file_name_template:
            render_dir_path = os.path.abspath(render_dir_path_template.render(tweet=tweet))
            os.makedirs(render_dir_path, exist_ok=True)
            render_file_name = render_file_name_template.render(tweet=tweet)
            render_file_path = os.path.join(render_dir_path, render_file_name)
            downloaded_image_file_paths = download_images(tweet, image_dir_path_template, image_url_path_template, render_file_name,
                                                          download_func, echo) if options['image-path'] else []

            if not options['append'] and os.path.exists(render_file_path):
                echo("Warning: Skipping existing file '{}'. Use --append to append rendered tweets instead.".format(render_file_path),
                     err=True)
                write_func = write_to_void
            else:
                write_func = write_to_file(render_file_path)
        else:
            downloaded_image_file_paths = []
            write_func = write_to_console

        rendered_tweet = template.render(tweet=tweet)

        if options['skip'] and options['skip'].search(rendered_tweet) is not None:
            echo("Warning: Skipping tweet ID {} ('{}{}') because its rendered form matches the --skip pattern."
                 .format(tweet.id, tweet.text[:30], "..." if len(tweet.text) > 30 else ""), err=True)
            cleanup_downloaded_images(downloaded_image_file_paths, echo)
            continue

        write_func(rendered_tweet)
        last_tweet_id = tweet.id

        # Check the count here as the list has already been "filtered" by this point and so the count remaining reflects the actual number
        # of tweets left to render.
        count_remaining -= 1
        pickle_tweets.append(tweet)

        if count_remaining == 0:
            break

    if not last_tweet_id:
        echo("Warning: No tweets were retrieved.", err=True)
    elif options['pickle']:
        try:
            with open(PICKLE_FILE_NAME, "wb") as pf:
                pickle.dump(pickle_tweets, pf)
        except OSError as e:
            echo("Warning: Unable to serialize the rendered tweets to '{}' in the current working directory: {}".format(PICKLE_FILE_NAME,
                                                                                                                        e))

    return last_tweet_id


def retrieve(auth_keys, options, template_text, echo):
    """ Using the given authorization credentials and rendering options, retrieve and render the tweets from the authorized user's timeline.
    """
    api = authenticate_twitter_api(**auth_keys)

    try:
        tweets = reversed(list(tweepy.Cursor(api.user_timeline, since_id=options['since-id'], include_rts=options['retweets']).items()))
    except tweepy.TweepError as e:
        raise TwempestError("Unable to retrieve tweets. Twitter API responded with '{}'. "
                            "See https://dev.twitter.com/overview/api/response-codes for an explanation.".format(e.response))

    return render(tweets, options, template_text, download_from_url, echo)



""" Twempest console application unit tests.
"""

# This file is part of Twempest. Copyright 2017 Dave Rogers <info@yukondude.com>. Licensed under the GNU General Public License, version 3.
# Refer to the attached LICENSE file or see <http://www.gnu.org/licenses/> for details.

import os
import re

from click.testing import CliRunner

from twempest import __version__
from twempest.__main__ import choose_config_path, choose_option_values, choose_since_id, last_tweet_id_file_name,\
    CONFIG_FILE_NAME, CONFIG_OPTIONS, twempest


def test_choose_config_path():
    with CliRunner().isolated_filesystem():
        os.makedirs("fff/ggg/hhh")
        os.makedirs("iii/jjj")
        os.makedirs("kkk")

        root = os.getcwd()
        os.chdir("fff/ggg")
        possible_paths = [os.path.abspath(os.path.join(root, p)) for p in ["kkk", "iii/jjj", "fff/ggg"]]

        assert choose_config_path(possible_paths[0], possible_paths[1], possible_paths[2], CONFIG_FILE_NAME)[0] is None

        for i in [2, 1, 0]:
            with open(os.path.join(possible_paths[i], CONFIG_FILE_NAME), 'w') as f:
                f.write("[twitter]")

            assert (possible_paths[i], possible_paths) == choose_config_path(possible_paths[0], possible_paths[1], possible_paths[2],
                                                                             CONFIG_FILE_NAME)

        for i in [2, 1, 0]:
            os.chmod(possible_paths[i], 0o555)

        assert (None, possible_paths) == choose_config_path(possible_paths[0], possible_paths[1], possible_paths[2], CONFIG_FILE_NAME)


def test_choose_option_values():
    class MockConfig(dict):
        def getboolean(self, k, d):
            return bool(self.get(k, d))

    cli_options = {'replies': True}
    config = MockConfig()
    config['retweets'] = True
    options = choose_option_values(CONFIG_OPTIONS, cli_options, config)
    assert options['render-path'] == CONFIG_OPTIONS['render-path'].default
    assert options['replies'] is True
    assert options['retweets'] is True


def test_choose_since_id():
    with CliRunner().isolated_filesystem():
        user_id = "xxx"

        since_id = choose_since_id("1234", user_id, ".")
        assert since_id == "1234"

        with open(last_tweet_id_file_name(user_id), 'w') as f:
            f.write("6789")

        since_id = choose_since_id(None, user_id, ".")
        assert since_id == "6789"

        since_id = choose_since_id("1234", user_id, ".")
        assert since_id == "1234"


HELP_OPTION_REGEX = re.compile(r"^Usage: twempest.+Show this message and exit\.$")


def test_help_option_switch():
    def version_option_switch(option_switch):
        runner = CliRunner()
        result = runner.invoke(twempest, [option_switch])
        assert result.exit_code == 0
        print(result.output)
        match = HELP_OPTION_REGEX.match(result.output.replace("\n", ""))
        assert match is not None

    version_option_switch("-h")
    version_option_switch("--help")


def test_last_tweet_id_file_name():
    fn1 = last_tweet_id_file_name("aaaaaaaaaaaaaaaaaaaaaaaaaaaaaa")
    fn2 = last_tweet_id_file_name("aaaaaaaaaaaaaaaaaaaaaaaaaaaaab")
    fn3 = last_tweet_id_file_name("aaaaaaaaaaaaaaaaaaaaaaaaaaaaaa")
    assert fn1 == fn3
    assert fn1 != fn2


def test_twempest_dry_run():
    with CliRunner().isolated_filesystem():
        runner = CliRunner()
        result = runner.invoke(twempest, [])
        assert result.exit_code != 0

        with open('template', 'w') as f:
            f.write("{{ tweet.text }}")

        with open(CONFIG_FILE_NAME, 'w') as f:
            f.write("[twempest]\n[twitter]\nconsumer_key=a\nconsumer_secret=b\naccess_token=c\naccess_token_secret=d")

        result = runner.invoke(twempest, ["-c", ".", "-s", "12345", "-D", "template"])
        assert "since-id = 12345" in result.output
        assert result.exit_code == 0


def test_twempest_fail_1_no_argument():
    with CliRunner().isolated_filesystem():
        runner = CliRunner()
        result = runner.invoke(twempest, [])
        assert result.exit_code != 0


def test_twempest_fail_2_missing_section():
    with CliRunner().isolated_filesystem():
        runner = CliRunner()
        result = runner.invoke(twempest, [])
        assert result.exit_code != 0

        with open('template', 'w') as f:
            f.write("{{ tweet.text }}")

        with open(CONFIG_FILE_NAME, 'w') as f:
            f.write("[twitter]")

        result = runner.invoke(twempest, ["-c", ".", "template"])
        assert "Could not find required '[twempest]' section" in result.output
        assert result.exit_code != 0


def test_twempest_fail_3_missing_auth():
    with CliRunner().isolated_filesystem():
        runner = CliRunner()
        result = runner.invoke(twempest, [])
        assert result.exit_code != 0

        with open('template', 'w') as f:
            f.write("{{ tweet.text }}")

        with open(CONFIG_FILE_NAME, 'w') as f:
            f.write("[twempest]\n[twitter]")

        result = runner.invoke(twempest, ["-c", ".", "template"])
        assert "Could not find required Twitter authentication credential 'consumer_key' in" in result.output
        assert result.exit_code != 0


def test_twempest_fail_4_count_not_integer():
    with CliRunner().isolated_filesystem():
        runner = CliRunner()
        result = runner.invoke(twempest, [])
        assert result.exit_code != 0

        with open('template', 'w') as f:
            f.write("{{ tweet.text }}")

        with open(CONFIG_FILE_NAME, 'w') as f:
            f.write("[twempest]\n[twitter]\nconsumer_key=a\nconsumer_secret=b\naccess_token=c\naccess_token_secret=d")

        result = runner.invoke(twempest, ["-c", ".", "-n", "'foo'", "template"])
        assert "The --count option must be an integer:" in result.output
        assert result.exit_code != 0


def test_twempest_fail_5_count_too_low():
    with CliRunner().isolated_filesystem():
        runner = CliRunner()
        result = runner.invoke(twempest, [])
        assert result.exit_code != 0

        with open('template', 'w') as f:
            f.write("{{ tweet.text }}")

        with open(CONFIG_FILE_NAME, 'w') as f:
            f.write("[twempest]\n[twitter]\nconsumer_key=a\nconsumer_secret=b\naccess_token=c\naccess_token_secret=d")

        result = runner.invoke(twempest, ["-c", ".", "-n", "0", "template"])
        assert "The --count option must be at least 1." in result.output
        assert result.exit_code != 0


def test_twempest_fail_6_image_path_no_render_file():
    with CliRunner().isolated_filesystem():
        runner = CliRunner()
        result = runner.invoke(twempest, [])
        assert result.exit_code != 0

        with open('template', 'w') as f:
            f.write("{{ tweet.text }}")

        with open(CONFIG_FILE_NAME, 'w') as f:
            f.write("[twempest]\n[twitter]\nconsumer_key=a\nconsumer_secret=b\naccess_token=c\naccess_token_secret=d")

        result = runner.invoke(twempest, ["-c", ".", "-i", ".", "template"])
        assert "Cannot download images unless the --render-file option is also specified." in result.output
        assert result.exit_code != 0


def test_twempest_fail_7_image_url_no_image_path():
    with CliRunner().isolated_filesystem():
        runner = CliRunner()
        result = runner.invoke(twempest, [])
        assert result.exit_code != 0

        with open('template', 'w') as f:
            f.write("{{ tweet.text }}")

        with open(CONFIG_FILE_NAME, 'w') as f:
            f.write("[twempest]\n[twitter]\nconsumer_key=a\nconsumer_secret=b\naccess_token=c\naccess_token_secret=d")

        result = runner.invoke(twempest, ["-c", ".", "-u", "foo/", "template"])
        assert "The --image-url option may only be specified if the --image-path option is as well." in result.output
        assert result.exit_code != 0


def test_twempest_fail_8_no_since_id():
    with CliRunner().isolated_filesystem():
        runner = CliRunner()
        result = runner.invoke(twempest, [])
        assert result.exit_code != 0

        with open('template', 'w') as f:
            f.write("{{ tweet.text }}")

        with open(CONFIG_FILE_NAME, 'w') as f:
            f.write("[twempest]\n[twitter]\nconsumer_key=a\nconsumer_secret=b\naccess_token=c\naccess_token_secret=d")

        result = runner.invoke(twempest, ["-c", ".", "template"])
        assert "The --since-id option is required" in result.output
        assert result.exit_code != 0


def test_twempest_fail_9_bad_skip_re():
    with CliRunner().isolated_filesystem():
        runner = CliRunner()
        result = runner.invoke(twempest, [])
        assert result.exit_code != 0

        with open('template', 'w') as f:
            f.write("{{ tweet.text }}")

        with open(CONFIG_FILE_NAME, 'w') as f:
            f.write("[twempest]\n[twitter]\nconsumer_key=a\nconsumer_secret=b\naccess_token=c\naccess_token_secret=d")

        result = runner.invoke(twempest, ["-c", ".", "-s", "12345", "-k", "'(foo'", "template"])
        assert "Syntax problem with --skip regular expression" in result.output
        assert result.exit_code != 0


def test_twempest_fail_10_response_401():
    with CliRunner().isolated_filesystem():
        runner = CliRunner()
        result = runner.invoke(twempest, [])
        assert result.exit_code != 0

        with open('template', 'w') as f:
            f.write("{{ tweet.text }}")

        with open(CONFIG_FILE_NAME, 'w') as f:
            f.write("[twempest]\n[twitter]\nconsumer_key=a\nconsumer_secret=b\naccess_token=c\naccess_token_secret=d")

        result = runner.invoke(twempest, ["-c", ".", "-s", "12345", "template"])
        assert "Response [401]" in result.output
        assert result.exit_code != 0


VERSION_OPTION_REGEX = re.compile(r"Twempest version (\d+\.\d+\.\d+)Copyright.+See LICENSE\.$")


def test_version_option_switch():
    def version_option_switch(option_switch):
        runner = CliRunner()
        result = runner.invoke(twempest, [option_switch])
        assert result.exit_code == 0
        match = VERSION_OPTION_REGEX.match(result.output.replace("\n", ""))
        assert match.group(1) == __version__

    version_option_switch("-V")
    version_option_switch("--version")



#!/usr/bin/env python
""" Twempest setuptools configuration.
"""

# This file is part of Twempest. Copyright 2017 Dave Rogers <info@yukondude.com>. Licensed under the GNU General Public License, version 3.
# Refer to the attached LICENSE file or see <http://www.gnu.org/licenses/> for details.

import os
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import sys

import twempest


if sys.version_info < (3, 3):
    sys.stderr.write("Twempest requires Python 3.3 or higher.\n")
    sys.exit(1)


HERE = os.path.abspath(os.path.dirname(__file__))


with open(os.path.join(HERE, "README.rst"), encoding='utf-8') as f:
    long_description = f.read().strip()


def gather_requirements(requirements_file_name):
    """ Return the list of required packages and versions from requirements.txt.
    """
    return [pkg.strip() for pkg in open(os.path.join(HERE, requirements_file_name), "r").readlines()]


# noinspection PyAttributeOutsideInit
class PyTest(TestCommand):
    test_args = ["--cov=twempest", "--cov=tests", "--cov-report=term-missing", "--cov-fail-under=80", "--flake8"]
    test_suite = True

    def finalize_options(self):
        TestCommand.finalize_options(self)

    def run_tests(self):
        # noinspection PyPackageRequirements
        import pytest
        sys.exit(pytest.main(self.test_args))


# noinspection PyAttributeOutsideInit
class PyCleanTest(PyTest):
    """ Same as PyTest, but clear the cache first.
    """
    test_args = ["--verbose", "--cache-clear", "--cov=twempest", "--cov=tests", "--cov-report=term-missing",
                 "--cov-fail-under=80", "--flake8"]


setup(
    author="Dave Rogers",
    author_email="thedude@yukondude.com",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Text Processing",
        "Topic :: Utilities",
    ],
    cmdclass={
        'test': PyTest,
        'cleantest': PyCleanTest,
    },
    description=twempest.__doc__.strip(),
    entry_points={
        'console_scripts': [
            "twempest=twempest.__main__:twempest",
        ]
    },
    extras_require={
        'testing': gather_requirements("requirements-test.txt"),
    },
    include_package_data=True,
    install_requires=gather_requirements("requirements.txt"),
    license="GPLv3",
    long_description=long_description,
    name="twempest",
    packages=find_packages(),
    platforms=["MacOS", "Linux"],
    tests_require=gather_requirements("requirements-test.txt"),
    url="https://github.com/yukondude/Twempest",
    version=twempest.__version__,
    zip_safe=False
)
